__all__ = [
    '__project__',
    '__title__',
    '__author__',
    '__copyright__',
    '__license__',
    '__version__',
    '__mail__',
    '__maintainer__',
    '__status__']

__project__ = 'PINA'
__title__ = "pina"
__author__ = "Nicola Demo, Maria Strazzullo"
__copyright__ = "Copyright 2021-2021, PINA contributors"
__license__ = "MIT"
__version__ = "0.0.1"
__mail__ = 'demo.nicola@gmail.com, '  # TODO
__maintainer__ = __author__
__status__ = "Alpha"
__packagename__ = "pina-mathlab"

